#!/usr/bin/perl
######################################################################
#
# $Id: nph-webjob.cgi,v 1.6 2003/02/20 14:08:11 mavrik Exp $
#
######################################################################
#
# Copyright 2001-2003 Klayton Monroe, Cable & Wireless
# All Rights Reserved.
#
######################################################################

use strict;

######################################################################
#
# NPH Specific Initialization
#
######################################################################

  my %returnCodes =
  (
    '200' => "OK",
    '250' => "Ping Received", # ftimes only
    '251' => "LinkTest Succeeded",
    '252' => "Upload Succeeded, Relay Failed", # ftimes only
    '405' => "Method Not Allowed",
    '450' => "Invalid Query",
    '451' => "File Already Exists",
    '452' => "Username Undefined",
    '453' => "Username-ClientId Mismatch",
    '454' => "Content-Length Undefined",
    '455' => "Content-Length Exceeds Limit",
    '456' => "Content-Length Mismatch",
    '457' => "File Not Available",
    '460' => "RequiredMask-FieldMask Mismatch", # ftimes only
    '500' => "Internal Server Error",
    '550' => "Insufficient Access"
  );

#######################################################################
#
# Platform Specific Initialization
#
#######################################################################

  my $baseDirectory = "/integrity";

  umask 022;

#######################################################################
#
# Site Specific Initialization
#
#######################################################################

  #####################################################################
  #
  # The $putNameFormat variable controls the location of uploaded
  # files in the $dropZone container. The default layout is flat
  # (i.e. everything goes in $dropZone). Alternate layouts can be
  # implemented by customizing the format string. A format string
  # is a combination of conversion specifications and plain text.
  # Supported conversion specifications include:
  #
  #   %CID = Client ID as a string
  #   %CMD = Command as a string
  #   %d   = Day of the month as a decimal number (01-31)
  #   %H   = Hour as a decimal number (00-23)
  #   %IP  = IP address as a dotted quad string
  #   %M   = Minute as a decimal number (00-59)
  #   %m   = Month as a decimal number (01-12)
  #   %PID = Process ID of server-side CGI script
  #   %S   = Second as a decimal number (00-60)
  #   %s   = Number of seconds since the Epoch
  #   %Y   = Year with century as a decimal number
  #
  # For example, the string:
  #
  #   "%CMD/%IP_%Y-%m-%d+%H:%M:%S"
  #
  # will cause uploaded files to be stored in sub-directories that
  # correspond to the name of the command executed, and each output
  # filename will consist of an IP address, date, and time.
  #
  # The added flexibility provided by this scheme means that it is
  # possible to create problematic format strings. Consider the
  # string shown here:
  #
  #   "%CID/%CMD"
  #
  # While this is a legal format string, it is likely to cause name
  # collisions. Therefore, it is important to create format strings
  # that contain enough job specific information to distinguish one
  # set of uploaded files from another.
  #
  #####################################################################

  my $putNameFormat   = "%CID_%Y%m%d%H%M%S_%CMD";

  #####################################################################
  #
  # The folder list array specifies additional locations where
  # shared programs/scripts may be found. If the requested file
  # does not exist in the default location (i.e. $clientId/commands),
  # the following locations are searched according to the order
  # given here.
  #
  #####################################################################

  my @folderList      = ("common", "shared");

  #####################################################################
  #
  # The require user variable forces the program to abort unless the
  # REMOTE_USER environment variable has been set. Apache sets this
  # REMOTE_USER when authentication is enabled.
  #
  #####################################################################

  my $requireUser     = 1; # 0 = disabled, 1 = enabled

  #####################################################################
  #
  # The require match variable forces the program to abort unless
  # $username matches $clientId. When this value is disabled, other
  # users may issue requests on behalf of a given client.
  #
  #####################################################################

  my $requireMatch    = 1; # 0 = disabled, 1 = enabled

  #####################################################################
  #
  # The enforce limit variable forces the program to abort whenever
  # $contentLength is greater than $maxLength.
  #
  #####################################################################

  my $enforceLimit    = 0; # 0 = disabled, 1 = enabled
  my $maxLength       = 50000000; # 50 MB

  #####################################################################
  #
  # The following regex variables are used to validate query strings.
  #
  #####################################################################

  my $reVersion       = "VERSION=(webjob[\\w\\d\. ]+)";
  my $reSystem        = "&SYSTEM=(.+)";
  my $reClientId      = "&CLIENTID=([\\w-]+)";
  my $reFilename      = "&FILENAME=([\\w\\.-]+)";
  my $reRunType       = "&RUNTYPE=(linktest|snapshot)";
  my $reOutLength     = "&STDOUT_LENGTH=(\\d{1,10})"; #4294967295
  my $reErrLength     = "&STDERR_LENGTH=(\\d{1,10})"; #4294967295
  my $reEnvLength     = "&STDENV_LENGTH=(\\d{1,10})"; #4294967295

#######################################################################
#
# Main Routine
#
#######################################################################

  ####################################################################
  #
  # Attempt to fire up the log file. Keep going if it fails.
  #
  ####################################################################

  open(LH, ">>$baseDirectory/logfiles/nph-webjob.log");

  ####################################################################
  #
  # These variables are included in every log entry.
  #
  ####################################################################

  my $logHandle       = \*LH;
  my $username        = $ENV{'REMOTE_USER'};
  my $cgiRemote       = $ENV{'REMOTE_ADDR'};
  my $cgiMethod       = $ENV{'REQUEST_METHOD'};
  my $cgiRequest      = undef;
  my $byteCount       = undef;
  my $startTime       = time;
  my $returnCode      = undef;
  my $context         = undef;
  my $explanation     = undef;

  ####################################################################
  #
  # Make sure that a username has been defined, if required.
  #
  ####################################################################

  if ($requireUser && (!defined $username || !length($username)))
  {
    $returnCode = '452';
    $context = undef;
    $explanation = $returnCodes{$returnCode};
    LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
    ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
  }

  ####################################################################
  #
  # Make sure that a content length has been defined.
  #
  ####################################################################

  my $contentLength = $ENV{'CONTENT_LENGTH'};

  if (!defined $contentLength || !length($contentLength))
  {
    $returnCode = '454';
    $context = undef;
    $explanation = $returnCodes{$returnCode};
    LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
    ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
  }

  ####################################################################
  #
  # Check that the content length is within the limit, if required.
  #
  ####################################################################

  if ($enforceLimit && $contentLength > $maxLength)
  {
    $returnCode = '455';
    $context = "Length=$contentLength Limit=$maxLength";
    $explanation = $returnCodes{$returnCode};
    LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
    ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
  }

  ####################################################################
  #
  # Make sure there is a writable drop zone.
  #
  ####################################################################

  my $dropZone = $baseDirectory . "/incoming";

  if (!-d $dropZone)
  {
    $returnCode = '500';
    $context = "DropZone=$dropZone";
    $explanation = $!;
    LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
    ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
  }

  if (!-W $dropZone)
  {
    $returnCode = '550';
    $context = "DropZone=$dropZone";
    $explanation = "Write access is required for this object";
    LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
    ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
  }

  ####################################################################
  #
  # Make sure there is a readable profiles directory.
  #
  ####################################################################

  my $profiles = $baseDirectory . "/profiles";

  if (!-d $profiles)
  {
    $returnCode = '500';
    $context = "Profiles=$profiles";
    $explanation = $!;
    LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
    ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
  }

  if (!-R $profiles)
  {
    $returnCode = '550';
    $context = "Profiles=$profiles";
    $explanation = "Read access is required for this object";
    LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
    ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
  }

  ####################################################################
  #
  # Preprocess the query string (i.e. unencode it).
  #
  ####################################################################

  my $queryString = $ENV{'QUERY_STRING'};
  $queryString =~ s/\+/ /g;
  $queryString =~ s/%([0-9a-fA-F]{2})/pack("c", hex($1))/ge;

  ####################################################################
  #
  # GET Request -- Determine what was requested, and serve it up.
  #
  ####################################################################

  if ($cgiMethod eq "GET")
  {
    my $reGet  = "^" . $reVersion . $reSystem . $reClientId . $reFilename . "\$";

    if ($queryString =~ /$reGet/)
    {
      my $version  = $1;
      my $system   = $2;
      my $clientId = $3;
      my $filename = $cgiRequest = $4;
      my $getFile  = undef;
      my $gotFile  = 0;

      if ($requireUser && $requireMatch && $username ne $clientId)
      {
        $returnCode = '453';
        $context = "Username=$username ClientId=$clientId";
        $explanation = $returnCodes{$returnCode};
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }

      foreach my $folder ($clientId, @folderList)
      {
        $getFile = $profiles . "/" . $folder . "/" . "commands" . "/" . $filename;
        if (-e $getFile)
        {
          $gotFile = 1;
          last;
        }
      }

      if ($gotFile)
      {
        if (open(FILE, "$getFile"))
        {
          $byteCount = -s $getFile;
          $returnCode = '200';
          ReturnContent(\*FILE, $byteCount, $returnCode, $returnCodes{$returnCode});
          $context = "File=$getFile";
          $explanation = $returnCodes{$returnCode};
          LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
          close(FILE);
          Finish($logHandle);
        }
        else
        {
          $byteCount = 0;
          $returnCode = '500';
          $context = "File=$getFile";
          $explanation = $!;
          LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
          ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
        }
      }
      else
      {
        $returnCode = '457';
        $context = "File=$getFile";
        $explanation = $returnCodes{$returnCode};
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }
    }
    else
    {
      $returnCode = '450';
      $context = "QueryString=" . $ENV{'QUERY_STRING'}; # Extract actual value from ENV.
      $explanation = $returnCodes{$returnCode};
      LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
      ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
    }
  }

  ####################################################################
  #
  # PUT Request -- Write the uploaded data directly to disk.
  #
  ####################################################################

  elsif ($cgiMethod eq "PUT")
  {
    my $rePut  = "^" . $reVersion . $reSystem . $reClientId . $reFilename . $reRunType . $reOutLength . $reErrLength . $reEnvLength . "\$";

    if ($queryString =~ /$rePut/)
    {
      my $version      = $1;
      my $system       = $2;
      my $clientId     = $3;
      my $filename     = $cgiRequest = $4;
      my $runType      = $5;
      my $stdoutLength = $6;
      my $stderrLength = $7;
      my $stdenvLength = $8;

      if ($requireUser && $requireMatch && $username ne $clientId)
      {
        $returnCode = '453';
        $context = "Username=$username ClientId=$clientId";
        $explanation = $returnCodes{$returnCode};
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }

      ##############################################################
      #
      # If this is a linktest, dump the data and return success.
      #
      ##############################################################

      if ($runType eq "linktest")
      {
        ReadAndChuckData($contentLength); # prevents a broken pipe
        $returnCode = '251';
        $context = undef;
        $explanation = $returnCodes{$returnCode};
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnSuccess($logHandle, $returnCode, $returnCodes{$returnCode});
      }

      ##############################################################
      #
      # Make filenames and directories.
      #
      ##############################################################

      my (%localError);

      my $putName = MakePutName($putNameFormat, $startTime, $clientId, $cgiRemote, $filename, \%localError);
      if (!defined $putName)
      {
        ReadAndChuckData($contentLength); # prevents a broken pipe
        $returnCode = '500';
        $context = $localError{'context'};
        $explanation = $localError{'message'};
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }
      my $outFile = $dropZone . "/" . $putName . ".out";
      my $errFile = $dropZone . "/" . $putName . ".err";
      my $envFile = $dropZone . "/" . $putName . ".env";
      my $rdyFile = $dropZone . "/" . $putName . ".rdy";

      if (!defined MakePutTree($dropZone, $putName, 0755, 1, \%localError))
      {
        ReadAndChuckData($contentLength); # prevents a broken pipe
        $returnCode = '500';
        $context = $localError{'context'};
        $explanation = $localError{'message'};
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }

      ##############################################################
      #
      # Make sure that none of the output files exist.
      #
      ##############################################################

      if (-e $outFile)
      {
        ReadAndChuckData($contentLength); # prevents a broken pipe
        $returnCode = '451';
        $context = "File=$outFile";
        $explanation = $returnCodes{$returnCode};
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }

      if (-e $errFile)
      {
        ReadAndChuckData($contentLength); # prevents a broken pipe
        $returnCode = '451';
        $context = "File=$errFile";
        $explanation = $returnCodes{$returnCode};
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }

      if (-e $envFile)
      {
        ReadAndChuckData($contentLength); # prevents a broken pipe
        $returnCode = '451';
        $context = "File=$envFile";
        $explanation = $returnCodes{$returnCode};
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }

      if (-e $rdyFile)
      {
        ReadAndChuckData($contentLength); # prevents a broken pipe
        $returnCode = '451';
        $context = "File=$rdyFile";
        $explanation = $returnCodes{$returnCode};
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }

      ##############################################################
      #
      # Initialize the read byte count.
      #
      ##############################################################

      $byteCount = 0;

      ##############################################################
      #
      # Suck in the out data and store it in a file. Abort, if the
      # file exists or can't be opened.
      #
      ##############################################################

      if (!open(FILE, ">$outFile"))
      {
        ReadAndChuckData($contentLength); # prevents a broken pipe
        $returnCode = '500';
        $context = "File=$outFile";
        $explanation = $!;
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }

      binmode(FILE);
      my $nOut = ReadAndWriteData(\*FILE, $stdoutLength);
      close(FILE);
      if ($nOut != $stdoutLength)
      {
        $returnCode = '456';
        $context = "File=$outFile NRead=$nOut Length=$stdoutLength";
        $explanation = $returnCodes{$returnCode};
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }
      $byteCount += $nOut;

      ##############################################################
      #
      # Suck in the err data and store it in a file. Abort, if the
      # file exists or can't be opened.
      #
      ##############################################################

      if (!open(FILE, ">$errFile"))
      {
        ReadAndChuckData($contentLength); # prevents a broken pipe
        $returnCode = '500';
        $context = "File=$errFile";
        $explanation = $!;
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }

      binmode(FILE);
      my $nErr = ReadAndWriteData(\*FILE, $stderrLength);
      close(FILE);
      if ($nErr != $stderrLength)
      {
        $returnCode = '456';
        $context = "File=$errFile NRead=$nErr Length=$stderrLength";
        $explanation = $returnCodes{$returnCode};
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }
      $byteCount += $nErr;

      ##############################################################
      #
      # Suck in the env data and store it in a file. Abort, if the
      # file exists or can't be opened.
      #
      ##############################################################

      if (!open(FILE, ">$envFile"))
      {
        ReadAndChuckData($contentLength); # prevents a broken pipe
        $returnCode = '500';
        $context = "File=$envFile";
        $explanation = $!;
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }

      binmode(FILE);
      my $nEnv = ReadAndWriteData(\*FILE, $stdenvLength);
      close(FILE);
      if ($nEnv != $stdenvLength)
      {
        $returnCode = '456';
        $context = "File=$envFile NRead=$nEnv Length=$stdenvLength";
        $explanation = $returnCodes{$returnCode};
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }
      $byteCount += $nEnv;

      ##############################################################
      #
      # Create the .rdy file. This file acts as a lock release. In
      # other words, its presence informs other applications that
      # it's safe to process this snapshot.
      #
      ##############################################################

      if (!open(FILE, ">$rdyFile"))
      {
        $returnCode = '500';
        $context = "File=$rdyFile";
        $explanation = $!;
        LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
        ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
      }
      print FILE "+";
      close(FILE);

      ####################################################################
      #
      # Return success.
      #
      ####################################################################

      $returnCode = '200';
      $context = undef;
      $explanation = $returnCodes{$returnCode};
      LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
      ReturnSuccess($logHandle, $returnCode, $returnCodes{$returnCode});
    }
    else
    {
      $returnCode = '450';
      $context = "QueryString=" . $ENV{'QUERY_STRING'}; # Extract actual value from ENV.
      $explanation = $returnCodes{$returnCode};
      LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
      ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
    }
  }

  ####################################################################
  #
  # Unsupported Request -- Return an error.
  #
  ####################################################################

  else
  {
    $returnCode = '405';
    $context = undef;
    $explanation = $returnCodes{$returnCode};
    LogMessage($logHandle, $username, $cgiRemote, $cgiMethod, $cgiRequest, $byteCount, $startTime, $returnCode, $context, $explanation);
    ReturnFailure($logHandle, $returnCode, $returnCodes{$returnCode});
  }

  1;


######################################################################
#
# Finish
#
######################################################################

sub Finish
{
  my ($logHandle) = @_;
  close($logHandle) if ($logHandle);
  exit;
}


######################################################################
#
# ReadAndWriteData
#
######################################################################

sub ReadAndWriteData
{
  my ($fileHandle, $toRead) = @_;
  my ($nRead, $nWritten);

  $nRead = $nWritten = 0;
  while($toRead)
  {
    my $data = "";
    $nRead = sysread(STDIN, $data, $toRead);
    $nWritten += syswrite($fileHandle, $data, $nRead);
    $toRead -= $nRead;
  }
  return $nWritten;
}


######################################################################
#
# ReadAndChuckData
#
######################################################################

sub ReadAndChuckData
{
  my ($toRead) = @_;

  while($toRead)
  {
    my $data = "";
    $toRead -= sysread(STDIN, $data, $toRead);
  }
  return $toRead;
}


######################################################################
#
# ReturnFailure
#
######################################################################

sub ReturnFailure
{
  my ($fileHandle, $statusCode, $reasonPhrase) = @_;
  print "HTTP/1.1 $statusCode $reasonPhrase\r\n";
  print "Server: $ENV{'SERVER_SOFTWARE'}\r\n";
  print "Content-Type: text/plain\r\n";
  print "Content-Length: 0\r\n";
  print "\r\n";
  Finish($fileHandle);
  1;
}


######################################################################
#
# ReturnSuccess
#
######################################################################

sub ReturnSuccess
{
  my ($fileHandle, $statusCode, $reasonPhrase) = @_;
  print "HTTP/1.1 $statusCode $reasonPhrase\r\n";
  print "Server: $ENV{'SERVER_SOFTWARE'}\r\n";
  print "Content-Type: text/plain\r\n";
  print "Content-Length: 0\r\n";
  print "\r\n";
  Finish($fileHandle);
  1;
}


######################################################################
#
# ReturnContent
#
######################################################################

sub ReturnContent
{
  my ($fileHandle, $contentLength, $statusCode, $reasonPhrase) = @_;
  print "HTTP/1.1 $statusCode $reasonPhrase\r\n";
  print "Server: $ENV{'SERVER_SOFTWARE'}\r\n";
  print "Content-Type: application/octet-stream\r\n";
  print "Content-Length: $contentLength\r\n";
  print "\r\n";
  while (<$fileHandle>)
  {
    print;
  }
  1;
}


######################################################################
#
# LogMessage
#
######################################################################

sub LogMessage
{
  my ($logHandle, $username, $address, $method, $request, $byteCount, $startTime, $status, $context, $explanation) = @_;
  my $time = (scalar localtime);
  my $runTime = time - $startTime;
  my $delimeter = " ";
  my @errorMessage;

  push(@errorMessage, ((defined $time)        ? $time               : "-"));
  push(@errorMessage, ((defined $username)    ? $username           : "-"));
  push(@errorMessage, ((defined $address)     ? $address            : "-"));
  push(@errorMessage, ((defined $method)      ? $method             : "-"));
  push(@errorMessage, ((defined $request)     ? $request            : "-"));
  push(@errorMessage, ((defined $byteCount)   ? $byteCount          : "-"));
  push(@errorMessage, ((defined $runTime)     ? $runTime            : "-"));
  push(@errorMessage, ((defined $status)      ? $status             : "-"));
  push(@errorMessage, ((defined $context)     ? "C{ $context }"     : "C{ - }"));
  push(@errorMessage, ((defined $explanation) ? "E{ $explanation }" : "E{ - }"));
  print $logHandle join($delimeter, @errorMessage) . "\n" if ($logHandle);
  1;
}


######################################################################
#
# MakePutName
#
######################################################################

sub MakePutName
{
  my ($putNameFormat, $time, $clientId, $clientIp, $command, $pError) = @_;

  my (
       %formatChecks,
       %formatValues,
       $outputString,
       $tokensString
     );

  if (
       !defined $putNameFormat ||
       !defined $time ||
       !defined $clientId ||
       !defined $clientIp ||
       !defined $command
     )
  {
    $$pError{'context'} = undef;
    $$pError{'message'} = "One or more inputs not defined.";
    return undef;
  }

  my (
       $second,
       $minute,
       $hour,
       $monthDay,
       $month,
       $year,
       $weekDay,
       $yearDay,
       $daylightSavings
     ) = gmtime($time);

  $formatValues{'CID'} = $clientId;
  $formatValues{'CMD'} = $command;
  $formatValues{'IP'}  = $clientIp;
  $formatValues{'Y'}   = sprintf("%04d", $year + 1900);
  $formatValues{'m'}   = sprintf("%02d", $month + 1);
  $formatValues{'d'}   = sprintf("%02d", $monthDay);
  $formatValues{'H'}   = sprintf("%02d", $hour);
  $formatValues{'M'}   = sprintf("%02d", $minute);
  $formatValues{'PID'} = sprintf("%d", $$);
  $formatValues{'S'}   = sprintf("%02d", $second);
  $formatValues{'s'}   = sprintf("%d", $time);

  $formatChecks{'CID'} = "[\\w-]+";
  $formatChecks{'CMD'} = "[\\w\\.-]+";
  $formatChecks{'IP'}  = "(\\d{1,3}\\.){3}\\d{1,3}";
  $formatChecks{'Y'}   = "\\d{4}";
  $formatChecks{'m'}   = "\\d{2}";
  $formatChecks{'d'}   = "\\d{2}";
  $formatChecks{'H'}   = "\\d{2}";
  $formatChecks{'M'}   = "\\d{2}";
  $formatChecks{'PID'} = "\\d+";
  $formatChecks{'S'}   = "\\d{2}";
  $formatChecks{'s'}   = "\\d+";

  foreach my $key (keys(%formatChecks))
  {
    if ($formatValues{$key} !~ /^$formatChecks{$key}$/)
    {
      $$pError{'context'} = "Key=$key Value=$formatValues{$key}";
      $$pError{'message'} = "Validation check failed.";
      return undef;
    }
  }

  $tokensString = join('|', sort(keys(%formatValues)));
  $outputString = $putNameFormat;
  $outputString =~ s/%($tokensString)/$formatValues{$1}/ge;

  return $outputString;
}


######################################################################
#
# MakePutTree
#
######################################################################

sub MakePutTree
{
  my ($dropZone, $putName, $mode, $popCount, $pError) = @_;
  my (@elements, $fullTree);

  $fullTree = $dropZone;
  @elements = split(/\//, $putName);

  while (defined $popCount && $popCount-- > 0)
  {
    pop(@elements);
  }

  foreach my $element (@elements)
  {
    $fullTree .= "/" . $element;
    if (!-d $fullTree && !mkdir($fullTree, $mode))
    {
      $$pError{'context'} = "Path=$fullTree";
      $$pError{'message'} = "$!";
      return undef;
    }
  }

  return 1;
}
